# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import pywst as pw


class RWST:
    """
    Reduced Wavelet Scattering Transform (RWST) class contains the output of the RWST of an image or a batch of images.
    This contains the corresponding coefficients and helps to handle and to plot these coefficients.
    """
    
    def __init__(self, J, L, model, loc_shape=(), j_min=0):
        """
        Constructor.

        Parameters
        ----------
        J : int
            Number of dyadic scales.
        L : int
            Number of angles between 0 and pi.
        model : RWSTModelBase
            Chosen RWST model object.
        loc_shape : tuple of ints, optional
            Shape of the local dimensions (batch dimension + local dimensions)
        j_min : type, optional
            Minimum dyadic scale. The default is 0.

        Returns
        -------
        None.

        """
        self.J = J
        self.L = L
        self.model = model
        self.j_min = j_min
        
        self.coeffs = {}
        self.coeffs_cov = {}
        
        # Information on the original WST coefficients
        self.wst_log2vals = False
        self.wst_normalized = False
        
        # Initialization for each of the three layer of coefficients
        self.coeffs['m0'] = np.zeros((1,) + loc_shape)
        self.coeffs['m1'] = np.zeros((J - j_min, model.layer1_nbparams + 1) + loc_shape)       # +1 to add chi2r coefficients
        self.coeffs['m2'] = np.zeros((J - j_min, J - j_min, model.layer2_nbparams + 1) + loc_shape)    # +1 to add chi2r coefficients
        self.coeffs_cov['m0'] = np.zeros((1, 1) + loc_shape) # Actually, it is not possible yet to get local information for this m0 covariance matrix.
        self.coeffs_cov['m1'] = np.zeros((J - j_min, model.layer1_nbparams, model.layer1_nbparams) + loc_shape)
        self.coeffs_cov['m2'] = np.zeros((J - j_min, J - j_min, model.layer2_nbparams, model.layer2_nbparams) + loc_shape)
        
    def _set_mask(self, mask):
        """
        Use MaskedArray objects to deal with masked values when applicable.

        Parameters
        ----------
        mask : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.coeffs['m0'] = ma.MaskedArray(self.coeffs['m0'])
        self.coeffs['m0'][:, mask] = ma.masked
        self.coeffs['m1'] = ma.MaskedArray(self.coeffs['m1'])
        self.coeffs['m1'][:, :, mask] = ma.masked
        self.coeffs['m2'] = ma.MaskedArray(self.coeffs['m2'])
        self.coeffs['m2'][:, :, :, mask] = ma.masked
        self.coeffs_cov['m0'] = ma.MaskedArray(self.coeffs_cov['m0'])
        if self.coeffs_cov['m0'][0, 0].shape == mask.shape: # Special case for m0 rwst coefficient as it is not possible for now to get local information.
            self.coeffs_cov['m0'][:, :, mask] = ma.masked
        self.coeffs_cov['m1'] = ma.MaskedArray(self.coeffs_cov['m1'])
        self.coeffs_cov['m1'][:, :, :, mask] = ma.masked
        self.coeffs_cov['m2'] = ma.MaskedArray(self.coeffs_cov['m2'])
        self.coeffs_cov['m2'][:, :, :, :, mask] = ma.masked

    def _set_coeffs(self, layer, jVals, coeffs, coeffs_cov, chi2r):
        """
        Internal functions. Set the values of the coefficients during the optimization (see RWSTOp.apply function).

        Parameters
        ----------
        layer : int
            Layer of coefficients.
        jVals : int or (int,int)
            If layer == 1, j_1 value. If layer == 2, (j_1,j_2) values. Otherwise, ignored.
        coeffs : array
            Coefficients found during the optimization.
        coeffs_cov : array
            Covariance matrix corresponding to the coefficients.
        chi2r : TYPE
            Reduced chi square values for the corresponding optimization.

        Returns
        -------
        None.

        """
        if layer == 0:
            self.coeffs['m0'] = coeffs
            self.coeffs_cov['m0'][:] = coeffs_cov
        elif layer == 1:
            self.coeffs['m1'][jVals - self.j_min, :-1] = coeffs
            self.coeffs['m1'][jVals - self.j_min, -1:] = chi2r
            self.coeffs_cov['m1'][jVals - self.j_min] = coeffs_cov
        elif layer == 2:
            j1, j2 = jVals
            self.coeffs['m2'][j1 - self.j_min, j2 -self.j_min, :-1] = coeffs
            self.coeffs['m2'][j1 - self.j_min, j2 - self.j_min, -1:] = chi2r
            self.coeffs_cov['m2'][j1 - self.j_min, j2 - self.j_min] = coeffs_cov

    def get_coeffs(self, name):
        """
        Returns the selected set of RWST coefficients.

        Parameters
        ----------
        name : str
            Can be 'S0' for layer 0 coefficients, 'chi2r1' or 'chi2r2' for reduced chi square values (for layer 1 and layer 2 optimization respectively),
            or the name of a term of the corresponding model (consistent with model.layer1_names and model.layer2_names).

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        array
            Corresponding coefficients.

        """
        if name == 'S0':
            return self.coeffs['m0']
        elif name == 'chi2r1':
            return self.coeffs['m1'][:, -1]
        elif name == 'chi2r2':
            return self.coeffs['m2'][:, :, -1]
        else:
            for index, nameList in enumerate(self.model.layer1_names):
                if name == nameList:
                    return self.coeffs['m1'][:, index]
            for index, nameList in enumerate(self.model.layer2_names):
                if name == nameList:
                    return self.coeffs['m2'][:, :, index]
        raise Exception("Unknown name of parameter: " + str(name))
        
    def get_coeffs_std(self, name):
        """
        Returns the standard deviation of the selected set of RWST coefficients.

        Parameters
        ----------
        name : str
            Can be 'S0' for layer 0 coefficients, or the name of a term of the corresponding model
            (consistent with model.layer1_names and model.layer2_names).

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        array
            Corresponding standard deviation values.

        """
        if name == 'S0':
            return np.sqrt(self.coeffs_cov['m0'][0, 0])
        else:
            for index, nameList in enumerate(self.model.layer1_names):
                if name == nameList:
                    return np.sqrt(self.coeffs_cov['m1'][:, index, index])
            for index, nameList in enumerate(self.model.layer2_names):
                if name == nameList:
                    return np.sqrt(self.coeffs_cov['m2'][:, :, index, index])
        raise Exception("Unknown name of parameter: " + str(name))
        
    def get_coeffs_cov(self, layer=None, j1=None, j2=None):
        """
        Returns the covariance matrix of the selected set of RWST coefficients.

        Parameters
        ----------
        layer : int, optional
            Layer of coefficients (must be 0, 1 or 2). The default is None.
        j1 : int, optional
            Specific j_1 value. The default is None.
        j2 : int, optional
            Specific j_2 value. The default is None.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        array
            Corresponding covariance matrix.

        """
        if layer == 0:
            return self.coeffs_cov['m0']
        elif layer == 1:
            if j1 is None:
                return self.coeffs_cov['m1']
            else:
                return self.coeffs_cov['m1'][j1 - self.j_min]
        elif layer == 2:
            if j1 is None and j2 is None:
                return self.coeffs_cov['m2']
            elif j1 is None and j2 is not None:
                return self.coeffs_cov['m2'][:, j2 - self.j_min]
            elif j1 is not None and j2 is None:
                return self.coeffs_cov['m2'][j1 - self.j_min, :]
            else:
                return self.coeffs_cov['m2'][j1 - self.j_min, j2 - self.j_min]
        else:
            raise Exception("Choose a layer between 0 and 2!")
            
    def _finalize(self):
        """
        Internal function that is called by RWSTOp.apply.
        Call the corresponding model finalize function that typically deal with potential model parameters degeneracies

        Returns
        -------
        None.

        """
        self.model.finalize(self)
            
    def _theta_labels(self, theta_range):
        """
        Internal function. theta_range must be an array of integers
        """
        ret = []
        for theta in theta_range:
            s = ""
            if theta == 0:
                s = "$0$"
            elif theta == self.L // 4:
                s = r"$\frac{\pi}{4}$"
            elif theta == -self.L // 4:
                s = r"$-\frac{\pi}{4}$"
            elif theta == self.L // 2:
                s = r"$\frac{\pi}{2}$"
            elif theta == -self.L // 2:
                s = r"$-\frac{\pi}{2}$"
            elif theta == 3 * (self.L // 4):
                s = r"$\frac{3\pi}{4}$"
            elif theta == - 3 * (self.L // 4):
                s = r"$-\frac{3\pi}{4}$"
            elif theta == self.L:
                s = r"$\pi$"
            elif theta == -self.L:
                s = r"$-\pi$"
            ret.append(s)
        return ret
            
    def plot(self, names=[], label=""):
        """
        Plot of the selected set of RWST coefficients.
        
        Default behaviour plots layer 1 and layer 2 coefficients, and the reduced chi square values.

        Parameters
        ----------
        names : str or list of str, optional
            List of coefficients we want to plot on the same figure.
            Can be "chi2r1" or "chi2r2" for reduced chi square values, or names included in model.layer1_nbparams and model.layer2_nbparams variables.
            The default is [].
        label : str, optional
            Label for the legend. The default is "".

        Returns
        -------
        None.

        """
        self.plot_compare([], names=names, labels=[label])
            
    def plot_compare(self, rwst_list, names=[], labels=[]):
        """
        Plot of the selected set of RWST coefficients of the current object next to those from a given list of other RWST objects.
        
        Default behaviour plots layer 1 and layer 2 coefficients, and the reduced chi square values.
        
        Note that the current object and the objects of rwst_list must be of consistent J, L, RWST model, j_min.

        Parameters
        ----------
        rwst_list : RWST or list of RWST
            RWST object or list of multiple RWST objects.
        names : str or list of str, optional
            List of coefficients we want to plot on the same figure.
            Can be "chi2r1" or "chi2r2" for reduced chi square values, or names included in model.layer1_nbparams and model.layer2_nbparams variables.
            The default is [].
        labels : list of str, optional
            List of labels to identify the current object and the input objects.
            If rwst_list is of length N, labels should be of length N+1 at most, where the first element refers to the label of the current object.
            Default label is "".

        Returns
        -------
        None.
        
        """
        # Check input
        if type(rwst_list) != list:
            rwst_list_loc = [rwst_list] # Easier to make a list in the following
        else:
            rwst_list_loc = rwst_list.copy() # Local shallow copy
        for elt in rwst_list_loc:
            if type(elt) != RWST:
                raise Exception("rwst_list must be a RWST object or a list of RWST objects!")
            else:
                if self.J != elt.J or self.L != elt.L or self.j_min != elt.j_min or type(self.model) != type(elt.model):
                    raise Exception("Inconsistent RWST objects.")
            
        if names == []:
            self.plot_compare(rwst_list, names=self.model.layer1_names, labels=labels)
            self.plot_compare(rwst_list, names=self.model.layer2_names, labels=labels)
            self.plot_compare(rwst_list, names=["chi2r1", "chi2r2"], labels=labels)
        else:
            rwst_list_loc = [self] + rwst_list_loc # Add current object to rwst_list_loc
            
            # Check labels
            if len(labels) < len(rwst_list_loc):
                labels += [""] * (len(rwst_list_loc) - len(labels)) # Fill up labels list to be consistent with the length of rwst_list
            
            # We first make sure to have a list of names for the following
            if type(names) != list:
                names = [names]
                
            indexLayer1 = []
            indexLayer2 = []
            for name in names:
                if name == "chi2r1":
                    indexLayer1.append(self.model.layer1_nbparams)
                elif name == "chi2r2":
                    indexLayer2.append(self.model.layer2_nbparams)
                for index, nameList in enumerate(self.model.layer1_names):
                    if name == nameList:
                        indexLayer1.append(index)
                for index, nameList in enumerate(self.model.layer2_names):
                    if name == nameList:
                        indexLayer2.append(index)
                        
            nRows = int(np.sqrt(len(indexLayer1) + len(indexLayer2)))
            nCols = int(np.ceil((len(indexLayer1) + len(indexLayer2)) / nRows))
            
            fig, axs = plt.subplots(nRows, nCols, figsize=(nCols * 4, nRows * 3))
            
            colorCycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
            
            for ax_index, pos in enumerate(np.ndindex((nRows, nCols))):
                legend = False # No legend except if there is a label to show in the following
                
                # Correction for weird behavior of subplots
                if nCols == 1 and nRows > 1:
                    pos = pos[0]
                elif nRows == 1 and nCols > 1:
                    pos = pos[1]
                elif nRows == 1 and nCols == 1:
                    axs = [axs]
                    pos = pos[0]
                
                color_cnt = 0 # Count for the color cycle
                
                for rwst_index, rwst_curr in enumerate(rwst_list_loc): # For every RWST object in rwst_list
                
                    # Get local parameters
                    loc_shape = rwst_curr.coeffs['m0'].shape[1:]
                    # Get a mask associated to S2 coefficient if any local coefficient turns out to be masked (we assume uniform mask along coefficient index axis).
                    mask = ma.getmaskarray(rwst_curr.coeffs['m2'][0,0,0])
                    # Define an index of local positions that are not masked.
                    validLocIndex = [loc for loc in np.ndindex(loc_shape) if mask[loc] == False]
                    
                    j1Vals = np.arange(rwst_curr.j_min, rwst_curr.J)
                    
                    if ax_index < len(indexLayer1): # Plot instructions for first order coefficients
                        index = indexLayer1[ax_index]
                        for locIndex in validLocIndex:
                            color = colorCycle[color_cnt % len(colorCycle)]
                            
                            coeffs = rwst_curr.coeffs['m1'][np.index_exp[:, index] + locIndex]
                            if index != rwst_curr.model.layer1_nbparams: # No std for chi2r1
                                coeffsStd = np.sqrt(rwst_curr.coeffs_cov['m1'][np.index_exp[:, index, index] + locIndex])
                            else:
                                coeffsStd = np.zeros(coeffs.shape)
                                
                            # Label design
                            label = labels[rwst_index]
                            if label != "" and locIndex != ():
                                label += " - "
                            if locIndex != ():
                                label += str(locIndex)
                            if label != "": legend = True # Need to display a legend
                                
                            axs[pos].plot(j1Vals, coeffs, label=label, color=color)
                            axs[pos].fill_between(j1Vals, coeffs - coeffsStd, coeffs + coeffsStd, alpha=0.2, color=color)
                            
                            color_cnt += 1

                        axs[pos].set_xlabel("$j_1$")
                        if index != rwst_curr.model.layer1_nbparams:
                            axs[pos].set_ylabel(rwst_curr.model.layer1_pltparams[index][0])
                            if rwst_curr.model.layer1_pltparams[index][1]: # Readable radian ylabels?
                                angleRange = np.arange((rwst_curr.coeffs['m1'][:, index].ravel().min() // (rwst_curr.L / 4)) * rwst_curr.L // 4, (rwst_curr.coeffs['m1'][:, index].ravel().max() // (rwst_curr.L / 4) + 2) * rwst_curr.L // 4, rwst_curr.L // 4)
                                axs[pos].set_yticks(angleRange)
                                axs[pos].set_yticklabels(rwst_curr._theta_labels(angleRange))
                        else:
                            axs[pos].set_ylabel(r"$\chi^{2, \mathrm{S_1}}_\mathrm{r}(j_1)$")
                            
                    elif ax_index < len(indexLayer1) + len(indexLayer2): # Plot instructions for second order coefficients
                        index = indexLayer2[ax_index - len(indexLayer1)]
                        for locIndex in validLocIndex:
                            color = colorCycle[color_cnt % len(colorCycle)]
                            
                            coeffs = rwst_curr.coeffs['m2'][np.index_exp[:, :, index] + locIndex]
                            if index != rwst_curr.model.layer2_nbparams: # No std for chi2r2
                                coeffsStd = np.sqrt(rwst_curr.coeffs_cov['m2'][np.index_exp[:, :, index, index] + locIndex])
                            else:
                                coeffsStd = np.zeros(coeffs.shape)
                            for j1 in np.arange(rwst_curr.j_min, rwst_curr.J - 1):
                                j2Vals = np.arange(j1 + 1, rwst_curr.J)
                                if j1 != rwst_curr.J - 2:
                                    
                                    # Label design
                                    label = ""
                                    if j1 == 0:
                                        label = labels[rwst_index]
                                        if label != "" and locIndex != ():
                                            label += " - "
                                        if locIndex != ():
                                            label += str(locIndex)
                                    if label != "": legend = True # Need to display a legend
                                        
                                    axs[pos].plot(j2Vals, coeffs[j1 - self.j_min, j1 - self.j_min + 1:], label=label, color=color)
                                    axs[pos].fill_between(j2Vals, coeffs[j1 - self.j_min, j1 - self.j_min + 1:] - coeffsStd[j1 - self.j_min,j1 - self.j_min + 1:], \
                                                          coeffs[j1 - self.j_min, j1 -self.j_min + 1:] + coeffsStd[j1 - self.j_min, j1 - self.j_min + 1:], alpha=0.2, color=color)
                                else:
                                    axs[pos].errorbar(j2Vals, coeffs[j1 - self.j_min, j1 - self.j_min + 1:], fmt='.', \
                                                      yerr=coeffsStd[j1 -self.j_min, j1 - self.j_min + 1:], color=color)
                                    
                            color_cnt += 1
                        
                        axs[pos].set_xlabel("$j_2$")
                        if index != rwst_curr.model.layer2_nbparams:
                            axs[pos].set_ylabel(rwst_curr.model.layer2_pltparams[index][0])
                            if rwst_curr.model.layer2_pltparams[index][1]: # Readable radian ylabels?
                                angleRange = np.arange((rwst_curr.coeffs['m2'][:, :, index].ravel().min() // (rwst_curr.L / 4)) * rwst_curr.L // 4, (rwst_curr.coeffs['m2'][:, :, index].ravel().max() // (rwst_curr.L / 4) + 2) * rwst_curr.L // 4, rwst_curr.L // 4)
                                axs[pos].set_yticks(angleRange)
                                axs[pos].set_yticklabels(rwst_curr._theta_labels(angleRange))
                        else:
                            axs[pos].set_ylabel(r"$\chi^{2, \mathrm{S_2}}_\mathrm{r}(j_1,j_2)$")
                        
                if legend:
                    axs[pos].legend()
                        
            plt.tight_layout()
            plt.show()
            
    def to_wst(self, cplx=False):
        """
        Return the corresponding WST object of the current object.

        Returns
        -------
        None.

        """
        # Build a WST index
        wst_index = [[0, 0, 0, 0, 0]] # Layer 0
        for j1 in range(self.j_min, self.J): # Layer 1
            for theta1 in range(self.L * (1 + cplx)):
                wst_index.append([1, j1, theta1, 0, 0])
        for j1 in range(self.j_min, self.J): # Layer 2
            for theta1 in range(self.L * (1 + cplx)):
                for j2 in range(j1 + 1, self.J):
                    for theta2 in range(self.L):
                        wst_index.append([2, j1, theta1, j2, theta2])
        wst_index = np.array(wst_index).T
        
        # Get local shape and mask
        loc_shape = self.coeffs['m0'][0].shape
        mask = ma.getmaskarray(self.coeffs['m2'][0,0,0])
        valid_loc_index = [loc for loc in np.ndindex(loc_shape) if mask[loc] == False]
        
        # Initialization
        S = np.zeros((wst_index.shape[1],) + loc_shape)
        if mask.sum() != 0: # We have at least one masked value
            S = ma.MaskedArray(S)
            S[:, mask] = ma.masked
        
        # S0 coefficients
        S[0] = self.coeffs['m0'][0]
        
        # S1 coefficients
        for j1 in range(self.j_min, self.J):
            filtering = np.logical_and(wst_index[0] == 1, wst_index[1] == j1)
            theta_vals = wst_index[2, filtering]
            coeffs = self.coeffs['m1'][j1 - self.j_min, :-1]
            for loc_index in np.ndindex(loc_shape):
                S[(filtering,) + loc_index] = self.model.layer1(theta_vals, *(coeffs[np.index_exp[:] + loc_index]))
            
        # S2 coefficients
        for j1 in range(self.j_min, self.J):
            for j2 in range(j1 + 1, self.J):
                filtering = np.logical_and(wst_index[1] == j1, wst_index[3] == j2)
                theta_vals = (wst_index[2, filtering], wst_index[4, filtering])
                coeffs = self.coeffs['m2'][j1 - self.j_min, j2 - self.j_min, :-1]
                for loc_index in valid_loc_index:
                    S[(filtering,) + loc_index] = self.model.layer2(theta_vals, *(coeffs[np.index_exp[:] + loc_index]))
                
        # We create the WST object and fill out its attributes
        wst = pw.WST(self.J, self.L, S, index=wst_index, cplx=cplx, j_min=self.j_min)
        wst.normalized = self.wst_normalized
        wst.log2vals = self.wst_log2vals
        
        # TODO: add uncertainties to wst
        
        return wst



# -*- coding: utf-8 -*-

import numpy as np
import scipy.optimize as opt
import numpy.linalg as la
import numpy.ma as ma
import warnings

from .wst import WST
from .wst_operator import WSTOp
from .rwst import RWST
from .rwst_models import RWSTModel1


class RWSTOp:
    """
    Reduced Wavelet Scattering Transform (RWST) operator.
    """
    
    def __init__(self, M, N, J, L=8, OS=0, cplx=False, model_cls=RWSTModel1, wst_op=None, j_min=0):
        """
        Constructor.
        
        Parameters
        ----------
        M : int
            Height of the input images.
        N : int
            Width of the input images.
        J : int
            Number of dyadic scales.
        L : int, optional
            Number of angles between 0 and pi. The default is 8.
        OS : int, optional
            Oversampling parameter. The default is 0.
        cplx : bool, optional
            Set it to true if the RWSTOp instance will ever apply to complex data. This would load in memory the whole set of bandpass filters. The default is False.
        model_cls : type, optional
            Class of the RWST model of the angular dependencies. The default is RWSTModel1.
        wst_op : WSTOp, optional
            WSTOp object with consistent parameters that avoids multiple filters loading.
        j_min : type, optional
            Minimum dyadic scale. The default is 0.

        Returns
        -------
        None.
        """
        self.M, self.N, self.J, self.L, self.OS, self.cplx, self.j_min = M, N, J, L, OS, cplx, j_min
        if wst_op is None:
            self.wst_op = WSTOp(M, N, J, L, OS, cplx, j_min=j_min)
        else:
            if wst_op.M == M and wst_op.N == N and wst_op.J == J and wst_op.L == L and wst_op.OS == OS and wst_op.cplx == cplx and wst_op.j_min == j_min:
                self.wst_op = wst_op
            else:
                warnings.warn("Warning! Loading WSTOp new instance because of wst_op inconsistencies.")
                self.wst_op = WSTOp(M, N, J, L, OS, cplx, j_min=j_min)
        self.model_cls = model_cls
    
    def apply(self, data, local=False, wst_average=False, crop=0.0, diag_cov=True):
        """
        Compute the RWST of input data or from a set of pre-computed WST coefficients.
        
        Input data can be either an image (2D array), or a batch of images (3D arrays, the first dimension is the image index) or a WST object.

        Parameters
        ----------
        data : WST or array
            WST object containing pre-computed WST coefficients or input data.
            2D (one image) or 3D array (batch of images).
        local : bool, optional
            Do we need local WST coefficients? Applicable if data is an image or a batch of images. The default is False.
        wst_average : bool, optional
            WST average before RWST computation? Applicable if data is an image or a batch of images. The default is False.
        crop : float, optional
            For non-periodic images, local coefficients at the borders may need to be cropped. Applicable if data is an image or a batch of images.
            Width of the cropping in 2^J pixels unit before downsampling (i.e. crop = 1 corresponds to 2^J pixels cropped before downsampling).
        diag_cov : bool, optional
            Only use the diagonal coefficients of the covariance matrix for curve_fit.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        rwst : RWST
            The RWST object that contains the corresponding RWST coefficients.

        """
        # Check if data is made of images or is directly a WST object.
        if type(data) == WST:
            wst = data
            if wst.J != self.J:
                raise Exception("Inconsistent J values between RWST operator and input WST coefficients.")
            if wst.j_min != self.j_min:
                raise Exception("Inconsistent j_min values between RWST operator and input WST coefficients.")
            if not wst.log2vals:
                warnings.warn("Input WST coefficients must have logarithmic values. Auto-computation of logarithmic coefficients...")
                wst.to_log2()
        else:
            wst = self.wst_op.apply(data, local, crop=crop)
            wst.to_log2()
            wst.normalize()
            if wst_average:
                wst.average()
        
        # Get the shape of the local information (batch + local coefficients per map)
        loc_shape = wst.coeffs.shape[1:]
        # Get a mask associated to S2 coefficient if any local coefficient turns out to be masked (we assume uniform mask along coefficient index axis).
        mask = ma.getmaskarray(wst.coeffs[-1])
        # Define an index of local positions that are not masked.
        validLocIndex = [loc for loc in np.ndindex(loc_shape) if mask[loc] == False]
        
        model = self.model_cls(self.L)
        rwst = RWST(self.J, self.L, model, loc_shape=loc_shape, j_min=self.j_min)
        
        # We keep relevant information on the WST object
        rwst.wst_log2vals = wst.log2vals
        rwst.wst_normalized = wst.normalized
        
        # Layer 0 coefficient
        coeffs, coeffsIndex = wst.get_coeffs(layer=0)
        coeffs_cov, _ = wst.get_coeffs_cov(layer=0)
        rwst._set_coeffs(0, None, coeffs, coeffs_cov, None)
        
        ##
        # Note on curve_fit/absolute_sigma parameter:
        # If coeffs_cov is None, we set absolute_sigma to False, otherwise it is True.
        # coeffs_cov absolute values are meaningful, this is why we need absolute_sigma=True,
        # but when coeffs_cov is None, the default value used by curve_fit is an identity matrix,
        # that is why we need absolute_sigma=False in this case.
        ##

        # Layer 1 coefficients
        for j1 in range(self.j_min, self.J):
            coeffs, coeffsIndex = wst.get_coeffs(layer=1, j1=j1)
            coeffs_cov, _ = wst.get_coeffs_cov(layer=1, j1=j1, autoremove_offdiag=not diag_cov)
            if diag_cov and coeffs_cov is not None: # Only keep diagonal coefficients if required
                coeffs_cov = np.diag(np.diag(coeffs_cov))
            theta1Vals = coeffsIndex[2]
            
            paramsOpt = np.zeros((model.layer1_nbparams,) + loc_shape)
            paramsCov = np.zeros((model.layer1_nbparams, model.layer1_nbparams) + loc_shape)
            chi2r = np.zeros(loc_shape)
            for locIndex in validLocIndex: # Enumerate the index values corresponding to the shape loc_shape
                # In the following, np.index_exp helps to concatenate slice() object and tuples.
                # We get the optimized parameters paramsOpt and their corresponding covariance matrix. paramsOpt minimizes the chi squared statistics.
                res = opt.curve_fit(model.layer1, theta1Vals, coeffs[np.index_exp[:] + locIndex],
                                    p0=np.ones(model.layer1_nbparams), sigma=coeffs_cov, absolute_sigma=coeffs_cov is not None, maxfev=5000)
                paramsOpt[np.index_exp[:] + locIndex], paramsCov[np.index_exp[:, :] + locIndex] = res
                x = coeffs[np.index_exp[:] + locIndex] - model.layer1(theta1Vals, *tuple(paramsOpt[np.index_exp[:] + locIndex]))
                if coeffs_cov is not None:
                    chi2r[locIndex] = x.T @ la.inv(coeffs_cov) @ x / (len(x) - model.layer1_nbparams)
                else:
                    chi2r[locIndex] = x.T @ x / (len(x) - model.layer1_nbparams)
            rwst._set_coeffs(1, j1, paramsOpt, paramsCov, chi2r)
            
        # Layer 2 coefficients
        for j1 in range(self.j_min, self.J):
            for j2 in range(j1 + 1, self.J):
                coeffs, coeffsIndex = wst.get_coeffs(layer=2, j1=j1, j2=j2)
                coeffs_cov, _ = wst.get_coeffs_cov(layer=2, j1=j1, j2=j2, autoremove_offdiag=not diag_cov)
                if diag_cov and coeffs_cov is not None: # Only keep diagonal coefficients if required
                    coeffs_cov = np.diag(np.diag(coeffs_cov))
                thetaVals = (coeffsIndex[2], coeffsIndex[4])
                
                paramsOpt = np.zeros((model.layer2_nbparams,) + loc_shape)
                paramsCov = np.zeros((model.layer2_nbparams, model.layer2_nbparams) + loc_shape)
                chi2r = np.zeros(loc_shape)
                for locIndex in validLocIndex: # Enumerate the index values corresponding to the shape loc_shape
                    # In the following, np.index_exp helps to concatenate slice() object and tuples.
                    # We get the optimized parameters paramsOpt and their corresponding covariance matrix. paramsOpt minimizes the chi squared statistics.
                    res = opt.curve_fit(model.layer2, thetaVals, coeffs[np.index_exp[:] + locIndex],
                                        p0=np.ones(model.layer2_nbparams), sigma=coeffs_cov, absolute_sigma=coeffs_cov is not None, maxfev=5000)
                    paramsOpt[np.index_exp[:] + locIndex], paramsCov[np.index_exp[:, :] + locIndex] = res
                    x = coeffs[np.index_exp[:] + locIndex] - model.layer2(thetaVals, *tuple(paramsOpt[np.index_exp[:] + locIndex]))
                    if coeffs_cov is not None:
                        chi2r[locIndex] = x.T @ la.inv(coeffs_cov) @ x / (len(x) - model.layer2_nbparams)
                    else:
                        chi2r[locIndex] = x.T @ x / (len(x) - model.layer1_nbparams)
                rwst._set_coeffs(2, (j1, j2), paramsOpt, paramsCov, chi2r)
        
        rwst._finalize() # Deal with potential model parameters degeneracies
        
        if mask.sum() != 0: # We have at least one masked value
            rwst._set_mask(mask) # Same mask as wst data
        
        return rwst



# -*- coding: utf-8 -*-
# This source file is inspired from the Kymatio project: https://www.kymat.io/index.html

import numpy as np
from scipy import ndimage


class Filter:
    """
        Base class for filters.
    """
    
    def __init__(self, M, N):
        self.M = M
        self.N = N
        self.data = np.zeros((M, N), complex)
        self.type = self.__class__.__name__
        
        
class GaborFilter (Filter):
    """
    Gabor filter.
    
    We make sure that the gabor filters have an (approximate?) unit L1-norm.
    """
    
    def __init__(self, M, N, j, theta, gamma=1.0, sigma0=1.0, k0=2*np.pi):
        """
        Constructor.

        Parameters
        ----------
        M : int
            Height.
        N : int
            Width.
        j : int
            Dyadic scale index.
        theta : float
            Rotation angle.
        gamma : float, optional
            Aspect ratio of the envelope. The default is 1.0.
        sigma0 : float, optional
            Standard deviation of the envelope before its dilation. The default is 1.0.
        k0 : float, optional
            Central wavenumber before the dilation. The default is 2 * np.pi.

        Returns
        -------
        None.

        """
        super().__init__(M, N)
        self.j = j
        self.theta = theta
        self.gamma = gamma
        self.sigma0 = sigma0
        self.sigma = self.sigma0 * 2 ** j
        self.k0 = k0
        self.k = self.k0 / 2 ** j
        self.build()
        
    def build(self):
        """
        Build the filter for a given set of parameters.
        """
        R = np.array([[np.cos(self.theta), -np.sin(self.theta)], [np.sin(self.theta), np.cos(self.theta)]])
        RInv = np.array([[np.cos(self.theta), np.sin(self.theta)], [-np.sin(self.theta), np.cos(self.theta)]])
        D = np.array([[1, 0], [0, self.gamma ** 2]])
        curv = np.dot(R, np.dot(D, RInv)) / (2 * self.sigma ** 2)
        
        for ex in [-2, -1, 0, 1]:
            for ey in [-2, -1, 0, 1]:
                [xx, yy] = np.mgrid[ex * self.M:self.M + ex * self.M, ey * self.N:self.N + ey * self.N]
                arg = -(curv[0, 0] * xx ** 2 + (curv[0, 1] + curv[1, 0]) * xx * yy + curv[1, 1] * yy ** 2) + 1.j * (xx * self.k * np.cos(self.theta) + yy * self.k * np.sin(self.theta))
                self.data += np.exp(arg)
                
        normFactor = 2 * np.pi * self.sigma ** 2 / self.gamma
        self.data /= normFactor
        
        
class GaussianFilter (Filter):
    """
    Gaussian filter.
    """
    
    def __init__(self, M, N, j, theta=0.0, gamma=1.0, sigma0=1.0):
        """
        Constructor.

        Parameters
        ----------
        M : int
            Height.
        N : int
            Width.
        j : int
            Dyadic scale index.
        theta : float, optional
            Rotation angle. The default is 0.0.
        gamma : float, optional
            Aspect ratio of the envelope. The default is 1.0.
        sigma0 : float, optional
            Standard deviation of the envelope before its dilation. The default is 1.0.

        Returns
        -------
        None.

        """
        super().__init__(M, N)
        self.data = np.zeros((M, N)) # No need for a complex data type
        self.j = j
        self.theta = theta
        self.gamma = gamma
        self.sigma0 = sigma0
        self.sigma = self.sigma0 * 2 ** j
        self.build()
        
    def build(self):
        """
        Build the filter for a given set of parameters.
        """
        R = np.array([[np.cos(self.theta), -np.sin(self.theta)], [np.sin(self.theta), np.cos(self.theta)]])
        RInv = np.array([[np.cos(self.theta), np.sin(self.theta)], [-np.sin(self.theta), np.cos(self.theta)]])
        D = np.array([[1, 0], [0, self.gamma ** 2]])
        curv = np.dot(R, np.dot(D, RInv)) / (2 * self.sigma ** 2)

        for ex in [-2, -1, 0, 1]:
            for ey in [-2, -1, 0, 1]:
                [xx, yy] = np.mgrid[ex * self.M:self.M + ex * self.M, ey * self.N:self.N + ey * self.N]
                arg = -(curv[0, 0] * xx ** 2 + (curv[0, 1] + curv[1, 0]) * xx * yy + curv[1, 1] * yy ** 2)
                self.data += np.exp(arg)
                
        normFactor = 2 * np.pi * self.sigma ** 2 / self.gamma
        self.data /= normFactor
        
        
class TopHatFilter (Filter):
    """
    Top-hat filter.
    """
    
    def __init__(self, M, N, j, theta=0.0, gamma=1.0, sigma0=1.0):
        """
        Constructor.

        Parameters
        ----------
        M : int
            Height.
        N : int
            Width.
        j : int
            Dyadic scale index.
        theta : float, optional
            Rotation angle. The default is 0.0.
        gamma : float, optional
            Aspect ratio of the envelope. The default is 1.0.
        sigma0 : float, optional
            Standard deviation of the envelope before its dilation. The default is 1.0.

        Returns
        -------
        None.

        """
        super().__init__(M, N)
        self.data = np.zeros((M, N)) # No need for a complex data type
        self.j = j
        self.theta = theta
        self.gamma = gamma
        self.sigma0 = sigma0
        self.sigma = self.sigma0 * 2 ** j
        self.build()
        
    def build(self):
        """
        Build the filter for a given set of parameters.
        """
        
        [xx, yy] = np.mgrid[0:self.M, 0:self.N]
        self.data = ((xx - self.M // 2) ** 2 + ((yy - self.N // 2) * self.gamma) ** 2 <= self.sigma ** 2).astype(float)
        
        self.data = np.fft.fftshift(ndimage.rotate(self.data, self.theta / np.pi * 180.0))
                
        self.data /= np.sum(self.data)
        
        
class Wavelet (Filter):
    """
    Base class for wavelets.
    """
    
    def __init__(self, M, N, j, theta):
        super().__init__(M, N)
        self.j = j
        self.theta = theta
        
        
class MorletWavelet (Wavelet):
    """
    Morlet Wavelet.
    
    We make sure that the wavelets have zero-mean and an (approximate?) unit L1-norm.
    """
    
    def __init__(self, M, N, j, theta, gamma=1.0, sigma0=1.0, k0=2*np.pi):
        """
        Constructor.

        Parameters
        ----------
        M : int
            Height.
        N : int
            Width.
        j : int
            Dyadic scale index.
        theta : float
            Rotation angle.
        gamma : float, optional
            Aspect ratio of the envelope. The default is 1.0.
        sigma0 : float, optional
            Standard deviation of the envelope before the dilation. The default is 1.0.
        k0 : float, optional
            Central wavenumber before the dilation. The default is 2 * np.pi.

        Returns
        -------
        None.

        """
        super().__init__(M, N, j, theta)
        self.gamma = gamma
        self.sigma0 = sigma0
        self.sigma = self.sigma0 * 2 ** j
        self.k0 = k0
        self.build()
        
    def build(self):
        """
        Build the filter for a given set of parameters.
        """
        gabor = GaborFilter(self.M, self.N, self.j, self.theta, self.gamma, self.sigma0, self.k0)
        gaussian = GaussianFilter(self.M, self.N, self.j, self.theta, self.gamma, self.sigma0)
        K = np.sum(gabor.data) / np.sum(gaussian.data)
        self.data = gabor.data - K * gaussian.data
        



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
import numpy as np


class RWSTModelBase (ABC):
    """
    Base class for the definition of a RWST model, i.e. a model of the angular dependencies of the WST coefficients.
    """

    def __init__(self, L):
        """
        Constructor.

        Parameters
        ----------
        L : int
            Number of angles between 0 and pi.

        Returns
        -------
        None.

        """
        self.L = L
        self.model = self.__class__.__name__
        self.layer1_names = []
        self.layer2_names = []
        self.layer1_pltparams = []
        self.layer2_pltparams = []
        self.layer1_nbparams = len(self.layer1_names)
        self.layer2_nbparams = len(self.layer2_names)
        
    def __str__(self):
        """
        String description of the model.

        Returns
        -------
        None.

        """
        s = "Model: " + type(self).__name__ + "\n"
        s += "--> Layer 0 coefficient: S0\n"
        s += "--> Layer 1 coefficients: "
        for name in self.layer1_names:
            s += name + " - "
        s = s[:-3] + "\n"
        s += "--> Layer 2 coefficients: "
        for name in self.layer2_names:
            s += name + " - "
        s = s[:-3]
        return s
        
    @abstractmethod
    def layer1(self, theta_vals, *params):
        """
        Model for layer 1 coefficients.

        Parameters
        ----------
        theta_vals : array
            theta_1 values.
        *params : float
            Parameters of the model.

        Returns
        -------
        array
            Predictions of the model.

        """
        theta1 = theta_vals
        pass
    
    @abstractmethod
    def layer2(self, theta_vals, *params):
        """
        Model for layer 2 coefficients.

        Parameters
        ----------
        theta_vals : (array, array)
            (theta_1,theta_2) values.
        *params : float
            Parameters of the model.

        Returns
        -------
        array
            Predictions of the model.
        """
        theta1, theta2 = theta_vals
        pass

    @abstractmethod
    def finalize(self, rwst):
        """
        Potential finalization steps on the RWST coefficients.

        Parameters
        ----------
        rwst : RWST
            RWST object to finalize.

        Returns
        -------
        None.

        """
        pass


class RWSTModel1 (RWSTModelBase):
    """
    Usual RWST model (without additional terms).
    """
    
    def __init__(self, L):
        super().__init__(L)
        self.layer1_names = ['S1Iso', 'S1Aniso', 'ThetaRef1']
        self.layer2_names = ['S2Iso1', 'S2Iso2', 'S2Aniso1', 'S2Aniso2', 'ThetaRef2']
        self.layer1_pltparams = [(r'$\hat{S}_1^{\mathrm{iso}}(j_1)$', False),
                                 (r'$\hat{S}_1^{\mathrm{aniso}}(j_1)$', False),
                                 (r'$\theta^{\mathrm{ref,1}}(j_1)$', True)]
        self.layer2_pltparams = [(r'$\hat{S}_2^{\mathrm{iso,1}}(j_1,j_2)$', False),
                                 (r'$\hat{S}_2^{\mathrm{iso,2}}(j_1,j_2)$', False),
                                 (r'$\hat{S}_2^{\mathrm{aniso,1}}(j_1,j_2)$', False),
                                 (r'$\hat{S}_2^{\mathrm{aniso,2}}(j_1,j_2)$', False),
                                 (r'$\theta^{\mathrm{ref,2}}(j_1,j_2)$', True)]
        self.layer1_nbparams = len(self.layer1_names)
        self.layer2_nbparams = len(self.layer2_names)
        
    def layer1(self, theta_vals, *params):
        theta1 = theta_vals
        return params[0] + params[1] * np.cos(2 * np.pi * (theta1 - params[2]) / self.L)
    
    def layer2(self, theta_vals, *params):
        theta1, theta2 = theta_vals
        return params[0] + params[1] * np.cos(2 * np.pi * (theta1 - theta2) / self.L) \
            + params[2] * np.cos(2 * np.pi * (theta1 - params[4]) / self.L) \
            + params[3] * np.cos(2 * np.pi * (theta2 - params[4]) / self.L)
    
    def finalize(self, rwst):
        locShape = rwst.coeffs['m0'].shape[1:]
        
        ###########
        # Layer 1 #
        ###########
        
        indexS1Aniso = self.layer1_names.index("S1Aniso")
        indexThetaRef1 = self.layer1_names.index("ThetaRef1")
        # Lift degeneracy between S1Aniso and ThetaRef1
        for j1 in range(rwst.j_min, rwst.J):
            filtering = rwst.coeffs['m1'][j1 - rwst.j_min, indexS1Aniso] < 0
            rwst.coeffs['m1'][j1 - rwst.j_min, indexS1Aniso, filtering] = np.abs(rwst.coeffs['m1'][j1 - rwst.j_min, indexS1Aniso, filtering])
            rwst.coeffs['m1'][j1 - rwst.j_min, indexThetaRef1, filtering] += self.L / 2
        # Smooth ThetaRef1 values
        rwst.coeffs['m1'][:, indexThetaRef1] = (rwst.coeffs['m1'][:, indexThetaRef1] + self.L / 2) % self.L - self.L / 2
        for j1 in range(rwst.j_min + 1, rwst.J):
            for locIndex in np.ndindex(locShape):
                if rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] - rwst.coeffs['m1'][(j1 - rwst.j_min - 1, indexThetaRef1) + locIndex] > self.L / 2:
                    rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] -= self.L
                elif rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] - rwst.coeffs['m1'][(j1 - rwst.j_min - 1, indexThetaRef1) + locIndex] < -self.L / 2:
                    rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] += self.L
        
        ###########
        # Layer 2 #
        ###########
        
        indexS2Aniso1 = self.layer2_names.index("S2Aniso1")
        indexS2Aniso2 = self.layer2_names.index("S2Aniso2")
        indexThetaRef2 = self.layer2_names.index("ThetaRef2")
        # Lift degeneracy between S2Aniso1 and ThetaRef2
        for j1 in range(rwst.j_min, rwst.J):
            for j2 in range(j1 + 1, rwst.J):
                filtering = rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso1] < 0
                rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso1, filtering] = np.abs(rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso1, filtering])
                rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso2, filtering] = - rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso2, filtering]
                rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2, filtering] += self.L / 2
        # Smooth ThetaRef2 values
        rwst.coeffs['m2'][:, :, indexThetaRef2] = (rwst.coeffs['m2'][:, :, indexThetaRef2] + self.L / 2) % self.L - self.L / 2
        for j1 in range(rwst.j_min, rwst.J):
            for j2 in range(j1 + 1, rwst.J):
                for locIndex in np.ndindex(locShape):
                    if rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] - rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min - 1, indexThetaRef2) + locIndex] > self.L / 2:
                        rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] -= self.L
                    elif rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] - rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min - 1, indexThetaRef2) + locIndex] < -self.L / 2:
                        rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] += self.L


class RWSTModel2 (RWSTModelBase):
    """
    RWST model that takes into account additional terms such as lattice terms and S2Iso3 term.
    """
    
    def __init__(self, L):
        super().__init__(L)
        self.layer1_names = ['S1Iso', 'S1Aniso', 'ThetaRef1', 'S1Lat1', 'S1Lat2']
        self.layer2_names = ['S2Iso1', 'S2Iso2', 'S2Aniso1', 'S2Aniso2', 'ThetaRef2', 'S2Iso3']
        self.layer1_pltparams = [(r'$\hat{S}_1^{\mathrm{iso}}(j_1)$', False),
                                 (r'$\hat{S}_1^{\mathrm{aniso}}(j_1)$', False),
                                 (r'$\theta^{\mathrm{ref,1}}(j_1)$', True),
                                 (r'$\hat{S}_1^{\mathrm{lat, 1}}(j_1)$', False),
                                 (r'$\hat{S}_1^{\mathrm{lat, 2}}(j_1)$', False)]
        self.layer2_pltparams = [(r'$\hat{S}_2^{\mathrm{iso,1}}(j_1,j_2)$', False),
                                 (r'$\hat{S}_2^{\mathrm{iso,2}}(j_1,j_2)$', False),
                                 (r'$\hat{S}_2^{\mathrm{aniso,1}}(j_1,j_2)$', False),
                                 (r'$\hat{S}_2^{\mathrm{aniso,2}}(j_1,j_2)$', False),
                                 (r'$\theta^{\mathrm{ref,2}}(j_1,j_2)$', True),
                                 (r'$\hat{S}_2^{\mathrm{iso,3}}(j_1,j_2)$', False)]
        self.layer1_nbparams = len(self.layer1_names)
        self.layer2_nbparams = len(self.layer2_names)
        
    def layer1(self, theta_vals, *params):
        theta1 = theta_vals
        return params[0] + params[1] * np.cos(2 * np.pi * (theta1 - params[2]) / self.L) \
            + params[3] * np.cos(4 * np.pi * theta1 / self.L) \
            + params[4] * np.cos(8 * np.pi * theta1 / self.L)
    
    def layer2(self, theta_vals, *params):
        theta1, theta2 = theta_vals
        return params[0] + params[1] * np.cos(2 * np.pi * (theta1 - theta2) / self.L) \
            + params[2] * np.cos(2 * np.pi * (theta1 - params[4]) / self.L) \
            + params[3] * np.cos(2 * np.pi * (theta2 - params[4]) / self.L) \
            + params[5] * np.cos(4 * np.pi * (theta1 - theta2) / self.L)
                
    def finalize(self, rwst):
        locShape = rwst.coeffs['m0'].shape[1:]
        
        ###########
        # Layer 1 #
        ###########
        
        indexS1Aniso = self.layer1_names.index("S1Aniso")
        indexThetaRef1 = self.layer1_names.index("ThetaRef1")
        # Lift degeneracy between S1Aniso and ThetaRef1
        for j1 in range(rwst.j_min, rwst.J):
            filtering = rwst.coeffs['m1'][j1 - rwst.j_min, indexS1Aniso] < 0
            rwst.coeffs['m1'][j1 - rwst.j_min, indexS1Aniso, filtering] = np.abs(rwst.coeffs['m1'][j1 - rwst.j_min, indexS1Aniso, filtering])
            rwst.coeffs['m1'][j1 - rwst.j_min, indexThetaRef1, filtering] += self.L / 2
        # Smooth ThetaRef1 values
        rwst.coeffs['m1'][:, indexThetaRef1] = (rwst.coeffs['m1'][:, indexThetaRef1] + self.L / 2) % self.L - self.L / 2
        for j1 in range(rwst.j_min + 1, rwst.J):
            for locIndex in np.ndindex(locShape):
                if rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] - rwst.coeffs['m1'][(j1 - rwst.j_min - 1, indexThetaRef1) + locIndex] > self.L / 2:
                    rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] -= self.L
                elif rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] - rwst.coeffs['m1'][(j1 - rwst.j_min - 1, indexThetaRef1) + locIndex] < -self.L / 2:
                    rwst.coeffs['m1'][(j1 - rwst.j_min, indexThetaRef1) + locIndex] += self.L
        
        ###########
        # Layer 2 #
        ###########
        
        indexS2Aniso1 = self.layer2_names.index("S2Aniso1")
        indexS2Aniso2 = self.layer2_names.index("S2Aniso2")
        indexThetaRef2 = self.layer2_names.index("ThetaRef2")
        # Lift degeneracy between S2Aniso1 and ThetaRef2
        for j1 in range(rwst.j_min, rwst.J):
            for j2 in range(j1 + 1, rwst.J):
                filtering = rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso1] < 0
                rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso1, filtering] = np.abs(rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso1, filtering])
                rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso2, filtering] = - rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexS2Aniso2, filtering]
                rwst.coeffs['m2'][j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2, filtering] += self.L / 2
        # Smooth ThetaRef2 values
        rwst.coeffs['m2'][:, :, indexThetaRef2] = (rwst.coeffs['m2'][:, :, indexThetaRef2] + self.L / 2) % self.L - self.L / 2
        for j1 in range(rwst.j_min, rwst.J):
            for j2 in range(j1 + 1, rwst.J):
                for locIndex in np.ndindex(locShape):
                    if rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] - rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min - 1, indexThetaRef2) + locIndex] > self.L / 2:
                        rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] -= self.L
                    elif rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] - rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min - 1, indexThetaRef2) + locIndex] < -self.L / 2:
                        rwst.coeffs['m2'][(j1 - rwst.j_min, j2 - rwst.j_min, indexThetaRef2) + locIndex] += self.L



# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import warnings
from .rwst import RWST


class WST:
    """
    Wavelet Scattering Transform (WST) class contains the output of the WST of an image or a batch of images.
    This contains the corresponding coefficients and helps to handle and to plot these coefficients.
    """
    
    def __init__(self, J, L, coeffs, index=None, cplx=False, j_min=0):
        """
        Constructor.

        Parameters
        ----------
        J : int
            Number of dyadic scales.
        L : int
            Number of angles between 0 and pi.
        coeffs : array
            WST coefficients. 1D, 2D, 3D, 4D arrays are valid depending whether
            we have local coefficients or not, whether we are dealing with the WST
            of a batch of images or not.
        index : array, optional
            Index of the coefficients stored in coeffs. The default is None.
        cplx : bool, optional
            Are these coefficients computed from complex images? The default is False.
        j_min : type, optional
            Minimum dyadic scale. The default is 0.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.J = J
        self.L = L
        self.cplx = cplx
        self.j_min = j_min
        
        # Number of coefficients
        self.nb_m0 = 1
        self.nb_m1 = (J - j_min) * L * (1 + cplx)
        self.nb_m2 = (J - j_min) * (J - j_min - 1) * L * L * (1 + cplx) // 2
        
        self.coeffs = coeffs
        self.coeffs_cov = None
        
        # Check if we are dealing with a batch of images coefficients or not, and if we have local coefficients or not.
        if coeffs.ndim == 1:
            self.batch = False
            self.local = False
        elif coeffs.ndim == 2:
            self.batch = True
            self.local = False
        elif coeffs.ndim == 3:
            self.batch = False
            self.local = True
        elif coeffs.ndim == 4:
            self.batch = True
            self.local = True
        else:
            raise Exception("Bad input shape.")
        
        # Reordering and definition of the current index
        if index is not None:
            self.reorder(index)
        self.index = np.zeros((5, self.nb_m0 + self.nb_m1 + self.nb_m2))
        cnt = 1
        for j1 in range(j_min, J):
            for theta1 in range(L * (1 + cplx)):
                self.index[:, cnt] = [1, j1, theta1, 0, 0]
                cnt += 1
        for j1 in range(j_min, J):
            for theta1 in range(L * (1 + cplx)):
                for j2 in range(j1 + 1, J):
                    for theta2 in range(L):
                        self.index[:, cnt] = [2, j1, theta1, j2, theta2]
                        cnt += 1
                        
        # Normalization initialization
        self.normalized = False
        self.normalize_layer1 = None
        self.log2vals = False
        
        # Plot design parameters
        self._xticks, self._xticksMinor, self._xticklabels, self._xticklabelsMinor, self._tickAlignment = [], [], [], [], []
        cnt = 1
        for j1 in range(j_min, J):
            self._xticks.append(cnt + L * (1 + cplx) // 2) # Tick at the middle
            self._xticklabels.append("$j_1 = " + str(j1) + "$")
            self._tickAlignment.append(0.0)
            cnt += L * (1 + cplx)
        for j1 in range(j_min, J - 1):
            for theta1 in range(L * (1 + cplx)):
                for j2 in range(j1 + 1, J):
                    self._xticksMinor.append(cnt + L // 2)
                    self._xticklabelsMinor.append("$j_2 = " + str(j2) + "$")
                    self._tickAlignment.append(-0.1)
                    cnt += L
            self._xticks.append(cnt - L * L * (1 + cplx) * (J - j1 - 1) // 2)
            self._xticklabels.append("$j_1 = " + str(j1) + "$")
            self._tickAlignment.append(0.0)
                
    def reorder(self, index):
        """
        Reorder coefficients (self.coeffs) to the standard lexicographical order (j1, theta1, j2, theta2) (ScatNet order).
        Index parameter specifies the current order.

        Parameters
        ----------
        index : array
            DESCRIPTION.

        Returns
        -------
        None.

        """
        coeffsCopy = self.coeffs.copy()
        
        # Lexicographical order
        reorderIndex = np.lexsort(index[::-1, :])

        # Reorder
        for i in range(coeffsCopy.shape[0]):
            self.coeffs[i, ...] = coeffsCopy[reorderIndex[i], ...]
                
    def _filter_args(self, layer=None, j1=None, theta1=None, j2=None, theta2=None):
        """
        Internal function to filter coefficients.

        Parameters
        ----------
        layer : int, optional
            Selection of the layer of coefficients (0, 1 or 2). The default is all layers.
        j1 : int, optional
            Selection of a specific j_1 values. The default is all j_1 values.
        theta1 : int, optional
            Selection of a specific theta_1 values. The default is all theta_1 values.
        j2 : int, optional
            Selection of a specific j_2 values. The default is all j_2 values.
        theta2 : int, optional
            Selection of a specific theta_2 values. The default is all theta_2 values.

        Returns
        -------
        filtering : array
            Boolean array for coefficients selection.

        """
        filtering = np.ones(self.index.shape[1], bool)
        if layer is not None:
            filtering = np.logical_and(filtering, self.index[0] == layer)
        if j1 is not None:
            filtering = np.logical_and(filtering, self.index[1] == j1)
        if theta1 is not None:
            filtering = np.logical_and(filtering, self.index[2] == theta1)
        if j2 is not None:
            filtering = np.logical_and(filtering, self.index[3] == j2)
        if theta2 is not None:
            filtering = np.logical_and(filtering, self.index[4] == theta2)
        return filtering
            
    def get_coeffs(self, layer=None, j1=None, theta1=None, j2=None, theta2=None):
        """
        Return the selected coefficients.

        Parameters
        ----------
        layer : int, optional
            Selection of the layer of coefficients (0, 1 or 2). The default is all layers.
        j1 : int, optional
            Selection of a specific j_1 values. The default is all j_1 values.
        theta1 : int, optional
            Selection of a specific theta_1 values. The default is all theta_1 values.
        j2 : int, optional
            Selection of a specific j_2 values. The default is all j_2 values.
        theta2 : int, optional
            Selection of a specific theta_2 values. The default is all theta_2 values.

        Returns
        -------
        array
            Array of the coefficients.
        array
            Index of the coefficients.

        """
        filtering = self._filter_args(layer=layer, j1=j1, theta1=theta1, j2=j2, theta2=theta2)
        return self.coeffs[filtering, ...], self.index[:, filtering]
    
    def to_log2(self):
        """
        Computation of logarithmic coefficients (binary logarithm).
        
        If coeffs_cov is not None, keep diagonal coefficients and divide them by (coeffs*log(2))^2 to estimate logarithmic errors at first order.

        Returns
        -------
        self

        """
        if not self.log2vals:
            # Compute the relevant covariance matrix, then compute log2 coefficients.
            if self.coeffs_cov is not None:
                warnings.warn("Warning! The covariance matrix has already been computed with linear coefficients. We compute logarithmic errors from diagonal coefficients and discard off-diagonal coefficients.")
                self.coeffs_cov = np.diag(np.diag(self.coeffs_cov)/(self.coeffs*np.log(2))**2)
            self.coeffs = ma.log2(self.coeffs)
            self.log2vals = True
        return self
            
    def to_linear(self):
        """
        Turn binary logarithmic coefficients into linear coefficients.
        
        If coeffs_cov is not None, keep diagonal coefficients and multiply them by (2^coeffs*log(2))^2 to estimate linear errors at first order.

        Returns
        -------
        self

        """
        if self.log2vals:
            # Compute linear coefficients, then compute the relevant covariance matrix.
            self.coeffs = 2 ** self.coeffs
            if self.coeffs_cov is not None:
                warnings.warn("Warning! The covariance matrix has already been computed with logarithmic coefficients. We compute linear errors from diagonal coefficients and discard off-diagonal coefficients.")
                self.coeffs_cov = np.diag(np.diag(self.coeffs_cov)*(self.coeffs*np.log(2))**2)
            self.log2vals = False
        return self
        
    def normalize(self, normalize_layer1=False):
        """
        Normalization of the coefficients.
        
        Layer 0 coefficients are left unchanged.
        Layer 1 coefficients are left unchanged if normalize_layer1 is False (default). Otherwise normalized by layer 0 coefficients (locally if available):
            
        .. math::
            
            \\bar{S}_1(j_1,\\theta_1) = S_1(j_1,\\theta_1)/S_0
            
        Layer 2 coefficients are normalized by the corresponding layer 1 coefficients (locally if available):
            
        .. math::
            
            \\bar{S}_2(j_1,\\theta_1,j_2,\\theta_2) = S_2(j_1,\\theta_1,j_2,\\theta_2)/S_1(j_1,\\theta_1)

        Parameters
        ----------
        normalize_layer1 : bool, optional
            Should layer 1 be also normalized? The default is False.

        Returns
        -------
        self

        """
        if self.coeffs_cov is not None:
            warnings.warn("Warning! The covariance matrix has been computed before the normalization and will not be updated.")
        
        # Case where no normalization was done or layer 1 normalization is asked to change
        if (not self.normalized) or (self.normalize_layer1 == (not normalize_layer1)):
            coeffsCopy = self.coeffs.copy()
            cnt = 1
            # Layer 1
            for j1 in range(self.j_min, self.J):
                # Case where layer 1 was not normalized but is now asked to be normalized
                if ((not self.normalized) or (not self.normalize_layer1)) and normalize_layer1:
                    if self.log2vals:
                        self.coeffs[cnt:cnt + self.L * (1 + self.cplx), ...] -= coeffsCopy[:1, ...]
                    else:
                        self.coeffs[cnt:cnt + self.L * (1 + self.cplx), ...] /= coeffsCopy[:1, ...]
                # Case where normalization was done with layer 1 normalized and layer 1 is asked not to be normalized
                elif self.normalized and self.normalize_layer1 and (not normalize_layer1):
                    if self.log2vals:
                        self.coeffs[cnt:cnt + self.L * (1 + self.cplx), ...] += coeffsCopy[:1, ...]
                    else:
                        self.coeffs[cnt:cnt + self.L * (1 + self.cplx), ...] *= coeffsCopy[:1, ...]
                cnt += self.L * (1 + self.cplx)
            # Layer 2
            if (not self.normalized):
                for j1 in range(self.j_min, self.J):
                    for theta1 in range(self.L * (1 + self.cplx)):
                        index = 1 + (j1 - self.j_min) * self.L * (1 + self.cplx) + theta1
                        if self.log2vals:
                            self.coeffs[cnt:cnt + self.L * (self.J - j1 - 1), ...] -= coeffsCopy[index:index + 1, ...]
                        else:
                            self.coeffs[cnt:cnt + self.L * (self.J - j1 - 1), ...] /= coeffsCopy[index:index + 1, ...]
                        cnt += self.L * (self.J - j1 - 1)
            self.normalized = True
            self.normalize_layer1 = normalize_layer1
        return self
            
    def unnormalize(self):
        """
        Cancel the normalization of the coefficients. See normalize() method.

        Returns
        -------
        self

        """
        if self.coeffs_cov is not None:
            warnings.warn("Warning! The covariance matrix has been computed before the unnormalization and will not be updated.")
        
        if self.normalized:
            cnt = 1
            for j1 in range(self.j_min, self.J):
                if self.normalize_layer1:
                    if self.log2vals:
                        self.coeffs[cnt:cnt + self.L * (1 + self.cplx), ...] += self.coeffs[:1, ...]
                    else:
                        self.coeffs[cnt:cnt + self.L * (1 + self.cplx), ...] *= self.coeffs[:1, ...]
                cnt += self.L * (1 + self.cplx)
            for j1 in range(self.j_min, self.J):
                for theta1 in range(self.L * (1 + self.cplx)):
                    index = 1 + (j1 - self.j_min) * self.L * (1 + self.cplx) + theta1
                    if self.log2vals:
                        self.coeffs[cnt:cnt + self.L * (self.J - j1 - 1), ...] += self.coeffs[index:index + 1, ...]
                    else:
                        self.coeffs[cnt:cnt + self.L * (self.J - j1 - 1), ...] *= self.coeffs[index:index + 1, ...]
                    cnt += self.L * (self.J - j1 - 1)
            self.normalized = False
            self.normalize_layer1 = None
        return self
                
    def average(self):
        """
        Computation of the mean coefficients and of the corresponding covariance matrix.
        
        This method needs batch data or local coefficients to be effective.

        Returns
        -------
        self

        """
        if self.batch or self.local: # We need multiple samples to average
            coeffsCopy = self.coeffs.copy()
            coeffsCopy = coeffsCopy.reshape((coeffsCopy.shape[0], -1))
            
            # How many valid samples do we have? (assumed to be the same for every coefficient)
            samples = 0
            if type(coeffsCopy) == ma.MaskedArray: # Do we have any masked value?
                samples = coeffsCopy[0].count(axis=-1) # We use S0 coefficient as it should be the same for any other coefficient
            else:
                samples = coeffsCopy.shape[-1]
                
            coeffs_cov = (ma.cov(coeffsCopy) / samples).data # Masked array covariance function to properly handle masked coefficients.
            
            self.coeffs = coeffsCopy.mean(axis=-1)
            self.coeffs_cov = coeffs_cov
            self.coeffs_cov_nbsamples = samples # Keep in memory the number of samples used to compute the sample covariance matrix.
            
            self.batch = False
            self.local = False
        return self
        
    def get_coeffs_cov(self, autoremove_offdiag=True, layer=None, j1=None, theta1=None, j2=None, theta2=None):
        """
        Return the covariance matrix corresponding to the selected coefficients.

        Parameters
        ----------
        layer : int, optional
            Selection of the layer of coefficients (0, 1 or 2). The default is all layers.
        j1 : int, optional
            Selection of a specific j_1 values. The default is all j_1 values.
        theta1 : int, optional
            Selection of a specific theta_1 values. The default is all theta_1 values.
        j2 : int, optional
            Selection of a specific j_2 values. The default is all j_2 values.
        theta2 : int, optional
            Selection of a specific theta_2 values. The default is all theta_2 values.
        autoremove_offdiag : bool, optional
            If we need to guarantee the covariance matrix to be definite, we may need to remove off diagonal coefficients.
            This is done when the effective number of samples to compute the sample covariance matrix is strictly lower than the dimension of the matrix plus one.
            The default is True.

        Returns
        -------
        array
            Covariance matrix corresponding to the selected coefficients.
        array
            Index of the selected coefficients.
        """
        filtering = self._filter_args(layer=layer, j1=j1, theta1=theta1, j2=j2, theta2=theta2)
        if self.coeffs_cov is None:
            return None, self.index[:, filtering]
        else:
            dim = np.sum(filtering)
            covM = self.coeffs_cov[np.outer(filtering, filtering)].reshape(dim, dim)
            # If demanded, remove off diagonal coefficients when we have too few samples to get a definite sample covariance matrix.
            if self.coeffs_cov_nbsamples < dim + 1 and autoremove_offdiag: # We typically expect definite sample covariance matrix when coeffs_cov_nbsamples >= dim + 1.
                covM = np.diag(np.diag(covM))
                warnings.warn("Warning! Removing off diagonal coefficients of the sample covariance matrix (only " + str(self.coeffs_cov_nbsamples) + " samples for dimension " + str(dim) + ").")
            return covM, self.index[:, filtering]
        
    def get_coeffs_std(self, layer=None, j1=None, theta1=None, j2=None, theta2=None):
        """
        Return the standard deviations corresponding to the selected coefficients.

        Parameters
        ----------
        layer : int, optional
            Selection of the layer of coefficients (0, 1 or 2). The default is all layers.
        j1 : int, optional
            Selection of a specific j_1 values. The default is all j_1 values.
        theta1 : int, optional
            Selection of a specific theta_1 values. The default is all theta_1 values.
        j2 : int, optional
            Selection of a specific j_2 values. The default is all j_2 values.
        theta2 : int, optional
            Selection of a specific theta_2 values. The default is all theta_2 values.

        Returns
        -------
        array
            Standard deviations corresponding to the selected coefficients.
        array
            Index of the selected coefficients.
        """
        cov, index = self.get_coeffs_cov(layer=layer, j1=j1, theta1=theta1, j2=j2, theta2=theta2)
        if cov is None:
            return None, index
        else:
            return np.sqrt(np.diag(cov)), index
        
    def _plot(self, axis, x, y, ylabel, legend="", err=None, j1ticks=True):
        """
        Internal plot function.

        Parameters
        ----------
        axis : matplotlib.axes.Axes
            Axes object.
        x : array
            x values.
        y : array
            y values.
        ylabel : str
            Label of y-axis.
        legend : str, optional
            Label for the legend. The default is "".
        err : array, optional
            Error on the coefficients. The default is None.
        j1ticks : bool, optional
            Do we want to show ticks for every j_1 value? The default is True.

        Returns
        -------
        None.

        """
        # Plot
        axis.plot(x, y, '-', markersize=2, label=legend)
        if err is not None:
            axis.fill_between(x, y - err, y + err, alpha=0.2)
        
        # Ticks and grid parameters
        axis.set_ylabel(ylabel)
        if j1ticks:
            axis.set_xticks(self._xticks)
            axis.set_xticklabels(self._xticklabels)
        else:
            axis.set_xticks([])
            axis.set_xticklabels([], visible=False)
        axis.set_xticks(self._xticksMinor, minor=True)
        axis.set_xticklabels(self._xticklabelsMinor, minor=True)
        for t, y in zip(axis.get_xticklabels(), self._tickAlignment):
            t.set_y(y)
        axis.grid(False, axis='x')
        axis.grid(False, axis='x', which='minor')
        axis.tick_params(axis='x', which='minor', direction='out', length=5)
        axis.tick_params(axis='x', which='major', bottom=False, top=False)
        axis.set_xlim(x.min() - 1, x.max() + 1)
        axis.margins(0.2)
        
        # Plot separators
        cnt = 1
        for j1 in range(self.j_min, self.J):
            axis.axvline(cnt, color='black', alpha=0.5, linestyle='dashed')
            cnt += self.L * (1 + self.cplx)
        for j1 in range(self.j_min, self.J):
            axis.axvline(cnt, color='black', alpha=0.5, linestyle='dashed')
            for theta1 in range(self.L * (1 + self.cplx)):
                if theta1 >= 1:
                    axis.axvline(cnt, color='black', alpha=0.2, linestyle='dashed')
                for j2 in range(j1 + 1, self.J):
                    if j2 > j1 + 1:
                        axis.axvline(cnt, color='black', alpha=0.1, linestyle=':')
                    cnt += self.L
        axis.axvline(cnt, color='black', alpha=0.5, linestyle='dashed')
            
        if legend != "":
            axis.legend()
        
    def plot(self, layer=None, j1=None, title="WST", label=""):
        """
        Plot the WST coefficients.
        Default behaviour plots both layer 1 and layer 2 full set of coefficients.
        
        Parameters
        ----------
        layer : int, optional
            Layer of coefficients to plot (1 or 2). The default is None.
        j1 : int, optional
            Selection of coefficients for a specific j_1 value. The default is None.
        title : str, optional
            Title of the figures. The default is "WST".
        label : str, optional
            Label for the legend. The default is "".

        Returns
        -------
        None.

        """
        self.plot_compare([], layer=layer, j1=j1, title=title, labels=[label])
                
    def plot_compare(self, wst_list, layer=None, j1=None, title="WST", labels=[]):
        """
        Plot of the selected set of WST coefficients of the current object next to those from a given list of other WST or RWST objects.
        
        Default behaviour plots both layer 1 and layer 2 full set of coefficients.
        
        Note that the current object and the objects of wst_list must be of consistent J, L, cplx and j_min parameters.

        Parameters
        ----------
        wst_list : (R)WST or list of (R)WST
            WST object or RWST object, or list of multiple WST/RWST objects.
        layer : int, optional
            Layer of coefficients to plot (1 or 2). The default is None.
        j1 : int, optional
            Selection of coefficients for a specific j_1 value. The default is None.
        title : str, optional
            Title of the figures. The default is "WST".
        labels : list of str, optional
            List of labels to identify the current object and the input objects.
            If wst_list is of length N, labels should be of length N+1 at most, where the first element refers to the label of the current object.
            Default label is "".

        Returns
        -------
        None.
        
        """
        # Check input
        if type(wst_list) != list:
            wst_list_loc = [wst_list] # Easier to make a list in the following
        else:
            wst_list_loc = wst_list.copy() # Local shallow copy

        for elt in wst_list_loc:
            if type(elt) != WST and type(elt) != RWST:
                raise Exception("wst_list must be a WST object or a RWST object, or a list of RWST or/and WST objects!")
            else:
                if self.J != elt.J or self.L != elt.L or self.j_min != elt.j_min or (type(elt) == WST and self.cplx != elt.cplx):
                    raise Exception("Inconsistent (R)WST objects.")
                if type(elt) == WST and (self.log2vals != elt.log2vals or self.normalized != elt.normalized):
                    raise warnings.warn("Warning! Input (R)WST objects have not consistent normalizations compared to the current WST object.")
            
        if layer is None:
            self.plot(layer=1, j1=j1, title=title)
            self.plot(layer=2, j1=j1, title=title)
        elif layer == 1 or layer == 2:
            wst_list_loc = [self] + wst_list_loc # Add current object to wst_list_loc
            
            # Check labels
            if len(labels) < len(wst_list_loc):
                labels += [""] * (len(wst_list_loc) - len(labels)) # Fill up labels list to be consistent with the length of wst_list
            
            # Selection of x values
            xValues = np.array(range(self.nb_m0 + self.nb_m1 + self.nb_m2))
            if j1 is None:
                xSelection = self._filter_args(layer=layer)
            else:
                xSelection = self._filter_args(layer=layer, j1=j1)
            
            # Figure/axis creation
            if layer == 1:
                fig = plt.figure()
            elif layer == 2:
                fig = plt.figure(figsize=(30, 5))
            ax = fig.add_subplot(1, 1, 1)
                
            # ylabel design
            ylabel = r"$" + self.log2vals * r"\log_2(" + self.normalized * (layer==2 or self.normalize_layer1==True) * r"\overline{" + r"S" + self.normalized * (layer==2 or self.normalize_layer1==True) * r"}" + r"_" + str(layer) + self.log2vals * r")" + r"$"
            
            # Plot
            for wst_index, wst_curr in enumerate(wst_list_loc):
                
                if type(wst_curr) == RWST:
                    wst_curr = wst_curr.to_wst(cplx=self.cplx) # If RWST object, convert it first to a WST object.
                
                # Get the shape of the local information (batch + local coefficients per map)
                loc_shape = wst_curr.coeffs.shape[1:]
                # Get a mask associated to S2 coefficient if any local coefficient turns out to be masked (we assume uniform mask along coefficient index axis).
                mask = ma.getmaskarray(wst_curr.coeffs[-1])
                # Define an index of local positions that are not masked.
                validLocIndex = [loc for loc in np.ndindex(loc_shape) if mask[loc] == False]
                
                for locIndex in validLocIndex:
                    # Label design
                    label = labels[wst_index]
                    if label != "" and locIndex != ():
                        label += " - "
                    if locIndex != ():
                        label += str(locIndex)
                    
                    coeffs = wst_curr.coeffs[tuple([xSelection]) + locIndex]
                    if wst_curr.coeffs_cov is None:
                        coeffsErr = None
                    else:
                        coeffsErr = np.sqrt(wst_curr.coeffs_cov.diagonal()[tuple([xSelection]) + locIndex])
                            
                    self._plot(ax, xValues[xSelection], coeffs, ylabel, legend=label, err=coeffsErr)
            
            # Finalization
            plt.subplots_adjust(bottom=0.2)
            plt.title(title + " - m = " + str(layer))
            if layer == 2:
                plt.tight_layout(rect=[0, 0.03, 1, 0.95])
            plt.show()



# -*- coding: utf-8 -*-

__all__ = ['wst_operator', 'wst', 'filters', 'rwst_operator', 'rwst', 'rwst_models']
__version__ = '1.0.1'

from .filters import *
from .wst import *
from .wst_operator import *
from .rwst import *
from .rwst_operator import *
from .rwst_models import *



# -*- coding: utf-8 -*-
# This source file is inspired from the Kymatio project: https://www.kymat.io/index.html

import os
import sys
from functools import partial
import warnings
import numpy as np
import numpy.ma as ma
import multiprocessing as mp
from .filters import MorletWavelet, GaussianFilter
from .utils import fft, ifft, subsample_fourier, modulus
from .wst import WST


# Internal function for the parallel pre-building of the bandpass filters (see WSTOp.load_filters)
def _build_bp_para(theta_list, bp_filter_cls, M, N, j, L, gamma, sigma0, k0):
    ret = []
    for theta in theta_list:
        ret.append(bp_filter_cls(M, N, j, (L // 2 - theta) * np.pi / L, gamma, sigma0, k0).data)
    return ret


class WSTOp:
    """
    Wavelet Scattering Transform (WST) operator.
    """
    
    def __init__(self, M, N, J, L=8, OS=0, cplx=False, lp_filter_cls=GaussianFilter, bp_filter_cls=MorletWavelet, j_min=0):
        """
        Constructor.
        
        M and N must be proportional to 2^J.

        Parameters
        ----------
        M : int
            Height of the input images.
        N : int
            Width of the input images.
        J : int
            Number of dyadic scales.
        L : int, optional
            Number of angles between 0 and pi. The default is 8.
        OS : int, optional
            Oversampling parameter. The default is 0.
        cplx : bool, optional
            Set it to true if the WSTOp instance will ever apply to complex data. This would load in memory the whole set of bandpass filters. The default is False.
        lp_filter_cls : type, optional
            Class corresponding to the low-pass filter. The default is GaussianFilter.
        bp_filter_cls : type, optional
            Class corresponding to the bandpass filter. The default is MorletWavelet.
        j_min : type, optional
            Minimum dyadic scale. The default is 0.

        Returns
        -------
        None.

        """
        if M % 2 ** J != 0 or N % 2 ** J != 0:
            raise Exception("Choose values for M and N that are proportional to 2^J.")
        
        self.M, self.N, self.J, self.L, self.OS, self.cplx, self.j_min = M, N, J, L, OS, cplx, j_min
        self.load_filters(lp_filter_cls, bp_filter_cls)
        
    def load_filters(self, lp_filter_cls, bp_filter_cls):
        """
        Build the set of low pass and bandpass filters that are used for the transform.

        Parameters
        ----------
        lp_filter_cls : type, optional
            Class corresponding to the low-pass filter.
        bp_filter_cls : type, optional
            Class corresponding to the bandpass filter.
            
        Returns
        -------
        None.

        """
        self.psi = {} # Bandpass filters
        self.phi = {} # Low-pass filters
        
        # Filter parameters
        gamma = 4 / self.L  # Aspect ratio
        sigma0 = 0.8        # Std of the envelope
        k0 = 3 * np.pi / 4  # Central wavenumber of the mother wavelet
        
        # Build psi filters
        for j in range(self.j_min, self.J):
            # Parallel pre-build
            build_bp_para_loc = partial(_build_bp_para, bp_filter_cls=bp_filter_cls, M=self.M, N=self.N, j=j, L=self.L, gamma=gamma, sigma0=sigma0, k0=k0)
            nb_processes = os.cpu_count()
            work_list = np.array_split(np.arange(self.L), nb_processes)
            pool = mp.Pool(processes=nb_processes)
            results = pool.map(build_bp_para_loc, work_list)
            bp_filters = []
            for i in range(len(results)):
                bp_filters += results[i]
            pool.close()
            
            for theta in range(self.L):
                self.psi[j, theta] = {}
                w = bp_filters[theta]
                wF = fft(w).real # The imaginary part is null for Morlet wavelets
                for res in range(max(j + 1, 1)):
                    self.psi[j, theta][res] = subsample_fourier(wF, 2 ** res, normalize=True, aa_filter=True)
            if self.cplx: # We also need rotations for theta in [pi, 2*pi)
                for theta in range(self.L):
                    self.psi[j, theta + self.L] = {}
                    for res in range(max(j + 1, 1)):
                        # Optimization trick for Morlet wavelets
                        self.psi[j, theta + self.L][res] = fft(np.conjugate(ifft(self.psi[j, theta][res]))).real
        
        # Build phi filters
        g = lp_filter_cls(self.M, self.N, self.J - 1, sigma0=sigma0).data
        gF = np.real(fft(g)) # The imaginary part is null for Gaussian filters
        for res in range(self.J):
            self.phi[res] = subsample_fourier(gF, 2 ** res, normalize=True, aa_filter=True)
    
    def apply(self, data, local=False, crop=0.0):
        """
        Compute the WST of input data.
        
        Input data can be either an image (2D array), or a batch of images (3D arrays, the first dimension is the image index).

        Parameters
        ----------
        data : array
            Input data. 2D (one image) or 3D array (batch of images).
        local : bool, optional
            Do we need local coefficients? The default is False.
        crop : float, optional
            For non-periodic images, local coefficients at the borders may need to be cropped.
            Width of the cropping in 2^J pixels unit before downsampling (i.e. crop = 1 corresponds to 2^J pixels cropped before downsampling).

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        WST
            The WST object that contains the corresponding WST coefficients.

        """
        
        J = self.J
        L = self.L
        OS = self.OS
        phi = self.phi
        psi = self.psi
        locCplx = self.cplx
        j_min = self.j_min
        
        # Check if we are dealing with a batch of images or not.
        if not (data.ndim == 2 or data.ndim == 3):
            raise Exception("Bad input shape.")
            
        # Check endianness for little-endian systems.
        # If input data is big-endian, we swap the byte order.
        # Should be useless with scipy 1.5.0.
        if sys.byteorder == 'little' and data.dtype.byteorder == '>':
            warnings.warn("Warning! Swapping byte order of input data to avoid FFT error.")
            data = data.byteswap().newbyteorder()
        
        if np.isrealobj(data):  # We make sure that data do not contain any complex value
            locCplx = False # We do not need the whole set of WST coefficients for real data
            
        # Check input images shape
        imgShape = (data.shape[-2], data.shape[-1])
        if imgShape[0] != self.M or imgShape[1] != self.N:
            raise Exception("Inconsistent width and height of the input images with the parameters of this operator.")

        # Output lists
        S = []
        SIndex = [] # Index of the coefficients of shape (5, nbCoeffs). Axis 0 order corresponds to: Layer, j1, theta1, j2, theta2
        
        # Zeroth layer
        U0 = fft(data)
        resU0 = 0
        if local:
            S0 = ifft(subsample_fourier(U0 * phi[0], 2 ** max(J - OS, 0)))
        else:
            S0 = ifft(U0)
        if locCplx:
            S0 = modulus(S0)
        else:
            S0 = S0.real # Discard the null imaginary part
        if local:
            S.append(S0)
        else:
            S.append(S0.mean(axis=(-1, -2)))
        SIndex.append([0, 0, 0, 0, 0])
    
        # First layer
        for j1 in range(j_min, J):
            for theta1 in range((locCplx + 1) * L): # Note the trick to produce 2 when locCplx == True, and 1 otherwise
                U1 = subsample_fourier(U0 * psi[j1, theta1][resU0], 2 ** max(j1 - resU0 - OS, 0))
                U1 = fft(modulus(ifft(U1)))
                resU1 = resU0 + max(j1 - resU0 - OS, 0)
                
                if local:
                    S1 = ifft(subsample_fourier(U1 * phi[resU1], 2 ** (max(J - resU1 - OS, 0)))).real
                    S.append(S1)
                else:
                    S1 = ifft(U1).real
                    S.append(S1.mean(axis=(-1, -2)))
                SIndex.append([1, j1, theta1, 0, 0])
            
                # Second layer
                for j2 in range(j1 + 1, J):
                    for theta2 in range(L):
                        U2 = subsample_fourier(U1 * psi[j2, theta2][resU1], 2 ** max(j2 - resU1 - OS, 0))
                        U2 = fft(modulus(ifft(U2)))
                        resU2 = resU1 + max(j2 - resU1 - OS, 0)
                        
                        if local:
                            S2 = ifft(subsample_fourier(U2 * phi[resU2], 2 ** (max(J - resU2 - OS, 0)))).real
                            S.append(S2)
                        else:
                            S2 = ifft(U2).real
                            S.append(S2.mean(axis=(-1, -2)))
                        SIndex.append([2, j1, theta1, j2, theta2])
                        
        S = np.array(S)
        
        # Cropping of the local coefficients at the border of the input images
        if local and crop != 0.0:
            mask = np.zeros(S.shape, bool)
            for i in range(self.M // 2 ** (J - OS)):
                for j in range(self.N // 2 ** (J - OS)):
                    if i < int(crop * 2 ** OS) or i > self.M // 2 ** (J - OS) - int(crop * 2 ** OS) \
                       or j < int(crop * 2 ** OS) or j > self.N // 2 ** (J - OS) - int(crop * 2 ** OS):
                        mask[..., i, j] = True
            if np.sum(~mask) == 0:
                raise Exception("No valid data remains after cropping.")
            S = ma.MaskedArray(S, mask=mask) # We create a numpy masked array to mask the cropped coefficients.
                        
        return WST(J, L, S, index=np.array(SIndex).T, cplx=locCplx, j_min=j_min)



# -*- coding: utf-8 -*-

import numpy as np
import scipy.fft


def fft(data):
    """
    Parallel FFT.
    """
    return scipy.fft.fft2(data, workers=-1) # By default, fft2 applies to the last two axes of data.


def ifft(data):
    """
    Parallel inverse FFT.
    """
    return scipy.fft.ifft2(data, workers=-1) # By default, ifft2 applies to the last two axes of data.


def subsample_fourier(x, k, normalize=False, aa_filter=False):
    """
    Periodization in Fourier space. This corresponds to a subsampling in physical space.

    Parameters
    ----------
    x : array
        Input data (2D or 3D arrays depending of a potential batch dimension).
    k : int
        DESCRIPTION.
    normalize : bool, optional
        Do we want to retain the L1 norm of the signal? The default is False.
    aa_filter : bool, optional
        Use an anti-aliasing filter before subsapmling? The default is False.

    Returns
    -------
    out : array
        Subsampled data (also in Fourier space).

    """
    # Anti-aliasing filter before subampling (if needed)
    if aa_filter:
        M, N = x.shape[-2], x.shape[-1]
        mask = np.ones((M, N))
        len_x = int(M * (1 - 1/k))
        len_y = int(N * (1 - 1/k))
        start_x = int(M / (2*k))
        start_y = int(N / (2*k))
        mask[start_x:start_x + len_x,:] = 0
        mask[:, start_y:start_y + len_y] = 0
        x = x * mask
    
    if x.ndim == 2: # Typical case: x is an image
        y = x.reshape(k, x.shape[0] // k, k, x.shape[1] // k)
        if normalize:
            out = y.sum(axis=(0, 2))
        else:
            out = y.mean(axis=(0, 2))
        return out
    elif x.ndim == 3: # x is a batch of images
        y = x.reshape(x.shape[0], k, x.shape[1] // k, k, x.shape[2] // k)
        if normalize:
            out = y.sum(axis=(1, 3))
        else:
            out = y.mean(axis=(1, 3))
        return out
    else:
        raise Exception("Bad input shape.")


def modulus(x):
    return np.absolute(x)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="pywst",
    version="1.0.1",
    author="Bruno Régaldo-Saint Blancard",
    author_email="bregaldo@flatironinstitute.org",
    description="WST and RWST analyses tools for astrophysical data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bregaldo/pywst",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
